/**
  *  \file client/dialogbinding.cc
  *  \brief Script Binding for Dialogs
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogbinding.h"
#include "client/keymaphandler.h"
#include "client/widgets/decaymsg.h"
#include "int/process.h"
#include "int/simple.h"
#include "ui/uiglobal.h"
#include "ui/window.h"
#include "util/atom.h"
#include "client/screens/minescreen.h"
#include "client/screens/ionscreen.h"
#include "client/screens/ufoscreen.h"
#include "game/global.h"

IntUserInterfaceDialogBinding::IntUserInterfaceDialogBinding()
    : IntUserInterfaceChildBinding(),
      eventQueue(),
      inLoop(false),
      keymap(0),
      keymap_prefix(0)
{ }

IntUserInterfaceDialogBinding::~IntUserInterfaceDialogBinding()
{ }

/** Main entry point. Call this instead of ui_root->run() to run the dialog.
    \return 0 if the dialog terminated with a iucExitDialog command or some script command,
    nonzero if it terminated by the C++ side executing stop(someNonzeroValue). */
int
IntUserInterfaceDialogBinding::runDialog()
{
    int result = 0;
    bool running = true;
    while (running) {
        if (!eventQueue.empty()) {
            // process queue, can terminate the screen
            std::auto_ptr<IntUserInterfaceEvent> event(eventQueue.extract_front());
            if (processEvent(event)) {
                running = false;
            }
        } else if (hasAnyRunnableProcess()) {
            // run pending processes
            runRunnableProcesses();
            killTerminatedProcesses();
            if (haveDisplayedTurn()) {
                getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
            }
        } else if (keymap != 0) {
            // pending UseKeymap
            doUseKeymap(keymap, keymap_prefix);
            keymap = 0;
        } else if (processPreIdle()) {
            // did something...
        } else {
            // user interaction
            inLoop = true;
            result = ui_root->run();
            inLoop = false;
            if (result != 0) {
                running = false;
            }
        }
    }
    return result;
}

/** Main entry point, shortcut for modal dialogs.
    This is a replacement for ui_root->runModal(w). */
int
IntUserInterfaceDialogBinding::runModal(UIBaseWidget& w)
{
    ui_root->add(w);
    int result = runDialog();
    ui_root->close(w);
    return result;
}

/** Check keymap command.
    \param key Key to check
    \param arg Prefix argument
    \retval true Command processed
    \retval false Key is not bound */
bool
IntUserInterfaceDialogBinding::checkKeymap(uint32_t key, int arg)
{
    std::auto_ptr<IntValue> v(getProperty(iuiKeymap));
    IntKeymapValue* kv = dynamic_cast<IntKeymapValue*>(v.get());
    if (kv != 0) {
        uint32_t cmd = kv->getKeymap()->lookupCommand(key);
        if (isAtom(cmd)) {
            executeStatement(atomStr(cmd), arg);
            triggerScripts();
            return true;
        } else {
            return false;
        }
    } else {
        return false;
    }
}

/** Trigger script execution. Call this whenever you assume there might be some
    runnable scripts awaiting execution. This will cause the window loop to exit
    and script execution to be triggered again. */
void
IntUserInterfaceDialogBinding::triggerScripts()
{
    // We want to generate just ONE stop, even if this is called multiple times.
    // Thus, just reset the "we are in a window loop" flag.
    if (inLoop) {
        inLoop = false;
        ui_root->stop(0);
    }
}

/** Process a command. Can be overridden to intercept specific commands.
    Always called from C++ context, not script context, and not the dialog's event loop.
    \retval true Exit the dialog
    \retval false Proceed with the dialog */
bool
IntUserInterfaceDialogBinding::processEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    switch (event->command) {
     case iucPopupConsole:
     case iucExitRace:
     case iucExitRaceNoSave:
     case iucExitClient:
     case iucGoToScreen:
     case iucGoToChart:
     case iucChartSetView:
        parent.postEvent(event);
        return true;

     case iucGoToMinefield:
        doMinefieldScreen();
        return false;

     case iucGoToUfo:
        doUfoScreen();
        return false;

     case iucGoToStorm:
        doIonStormScreen();
        return false;

     case iucOverlayMessage:
        showDecayingMessage(IntValue::toStringOf(event->args[0], false));
        return false;

     case iucExitDialog:
        return true;

     case iucUseKeymap:
        if (IntKeymapValue* kv = dynamic_cast<IntKeymapValue*>(event->args[0])) {
            IntIntValue* iv = dynamic_cast<IntIntValue*>(event->args[1]);
            keymap = kv->getKeymap();
            keymap_prefix = iv ? iv->getValue() : 0;
        }
        return false;
    }
    return false;
}

/** Perform processing before going to user-interface interaction.
    By default, does nothing.
    Can be overridden to perform specific tasks before the dialog (re)enters its event loop.
    \retval true Function did something, re-check commands etc.
    \retval false Function did nothing, enter event loop */
bool
IntUserInterfaceDialogBinding::processPreIdle()
{
    return false;
}

// Implementation of IntUserInterfaceChildBinding methods:
IntValue*
IntUserInterfaceDialogBinding::getProperty(IntUserInterfaceProperty p)
{
    return parent.getProperty(p);
}

void
IntUserInterfaceDialogBinding::setProperty(IntUserInterfaceProperty p, IntValue* v)
{
    parent.setProperty(p, v);
}

void
IntUserInterfaceDialogBinding::postEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    eventQueue.push_back(event.release());
    triggerScripts();
}

void
IntUserInterfaceDialogBinding::enumContexts(ptr_vector<IntContext>& ctx)
{
    parent.enumContexts(ctx);
}
