/**
  *  \file client/dialogs/attachmentselect.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/attachmentselect.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/attachmentunpacker.h"
#include "gfx/palette.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/quit.h"
#include "util/nls.h"

namespace {
    const int CB_SIZE = 24;

    int getWidth()
    {
        return 20*font_ems[FONT_NORMAL];
    }

    class AttachmentList : public UIListbox {
     public:
        AttachmentList(GAttachmentUnpacker& unp)
            : UIListbox(0, std::max(3, std::min(10, int(unp.getNumAttachments()))), getWidth(), 0),
              m_unpacker(unp)
            {
                setItemHeight(std::max(font_heights[FONT_NORMAL] + font_heights[FONT_SMALL], CB_SIZE));
                setSize(unp.getNumAttachments());
            }

        void drawPart(GfxCanvas& can, int from, int to)
            {
                GfxContext ctx;
                ctx.useCanvas(can);
                for (int i = from; i < to; ++i) {
                    GfxRect area;
                    prepareEntry(ctx, ColorStyle, i, &area);

                    if (GAttachmentUnpacker::Attachment* a = m_unpacker.getAttachmentByIndex(i + getTop())) {
                        // Checkbox
                        GfxRect cbArea = area.splitX(CB_SIZE);
                        Ptr<GfxPixmap> cb = requestPixmap(ResId("cb", m_unpacker.isAttachmentSelected(a) ? 2 : 1));
                        if (cb.ptr() != 0) {
                            cb->blit(can, cbArea.x + 12 - cb->getWidth()/2, cbArea.y + 12 - cb->getHeight()/2);
                        }

                        // Name & Information
                        const int TOP = font_heights[FONT_NORMAL];
                        const int BOT = font_heights[FONT_SMALL];
                        int height = TOP + BOT;
                        if (area.h > height) {
                            area.consumeY((area.h - height) / 2);
                        }
                        ctx.setColor(m_unpacker.getAttachmentKind(a) == GAttachmentUnpacker::CriticalFile ? UIColor::tc_Red : UIColor::tc_Static);
                        ctx.setTextAlign(0, 0);
                        ctx.useFont(*fonts[FONT_NORMAL]);
                        outTextF(ctx, area.splitY(TOP), m_unpacker.getAttachmentName(a));
                        ctx.useFont(*fonts[FONT_SMALL]);
                        outTextF(ctx, area, format(_("(%s, %d byte%!1{s%})"), GAttachmentUnpacker::toString(m_unpacker.getAttachmentKind(a)), m_unpacker.getAttachmentSize(a)));
                    }
                }
            }

        void toggleSelection()
            {
                GAttachmentUnpacker::Attachment* a = m_unpacker.getAttachmentByIndex(getCurrent());
                m_unpacker.selectAttachment(a, !m_unpacker.isAttachmentSelected(a));
                drawWidget();
            }

        void onItemClick(int /*item*/, int x)
            {
                if (x < CB_SIZE) {
                    toggleSelection();
                }
            }

     private:
        GAttachmentUnpacker& m_unpacker;
    };

    class AttachmentKeyHandler : public UIFixedWidget {
     public:
        AttachmentKeyHandler(AttachmentList& list)
            : m_list(list)
            { }
        bool handleEvent(const UIEvent& event, bool second_pass)
            {
                if (event.type == ev_Key) {
                    if (event.key.code == ' ') {
                        m_list.toggleSelection();
                        return true;
                    }
                }
                return UIFixedWidget::handleEvent(event, second_pass);
            }

        void drawContent(GfxCanvas& /*can*/)
            { }

     private:
        AttachmentList& m_list;
    };

    class AttachmentDialog {
     public:
        AttachmentDialog(GAttachmentUnpacker& unp)
            : m_list(unp)
            { }

        bool run()
            {
                ObjectHolder del;
                UIWindow win(_("Attachments"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
                win.setLayout(UIVBoxLayout::instance5);

                win.add(del.add(new UIMultilineStatic(FONT_NORMAL, _("You have received some additional files with your result. Choose which files you want to accept:"), getWidth())));

                win.add(UIFrameGroup::wrap(del, m_list));

                UIWidget& keys = del.add(new AttachmentKeyHandler(m_list));
                win.add(keys);

                UIButton& btnOK = del.add(new UIButton(1, _("OK"), SDLK_RETURN));
                UIButton& btnCancel = del.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE));
                UIGroup& btns = del.add(new UIGroup(UIHBoxLayout::instance5));
                btns.add(btnOK);
                btns.add(btnCancel);
                btns.add(del.add(new UISpacer()));
                btns.add(del.add(new UIButton(_("Help"), 'h')));
                win.add(btns);
                btnOK.addStop();
                btnCancel.addStop();

                win.add(del.add(new UIQuit(0)));
                win.add(del.add(new WHelpWidget("pcc2:resultattachments")));

                win.pack();
                ui_root->centerWidget(win);
                return ui_root->runModal(win);
            }

     private:
        AttachmentList m_list;
    };
}

bool chooseAttachments(GAttachmentUnpacker& unp)
{
    if (unp.getNumAttachments() != 0) {
        return AttachmentDialog(unp).run();
    } else {
        return false;
    }
}
