/**
  *  \file flak/flak.h
  *  \brief Fleet Action Kombat
  *  \ingroup FLAK
  *
  *  This contains the main definitions for Fleet Action Kombat, a
  *  multi-ship combat system jointly developed by the Planetsserver /
  *  Vagabond / PHost crew.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FLAK_FLAK_H
#define PCC_V2_FLAK_FLAK_H

#include <vector>
#include <ostream>
#include "flak/flakdefs.h"
#include "cpluslib/buffer.h"

class FlakFleet;
class FlakBattle;
class Stream;
class FlakVisualizer;

extern const char FLAK_MAGIC[8];

/** Position in 3D space. */
struct FlakPos {
    int32 x, y, z;

    FlakPos()
        { }
    FlakPos(int32 x, int32 y, int32 z)
        : x(x), y(y), z(z) { }

    double distanceTo(const FlakPos& other) const;
    bool   isDistanceLERadius(const FlakPos& other, int32 radius) const;
};

class FlakShip {
 public:
    /** Running status for this ship, variable. */
    struct Status {
        double  shield;
        double  damage;
        double  crew;
        bool    is_alive;
        const FlakShip* last_hit_by;
        int     beam_charge[FLAK_MAX_BEAMS];
        int     torp_charge[FLAK_MAX_TORPS];
        int     bay_charge[FLAK_MAX_BAYS];
        int     torp_limit;
        int     launch_countdown;
        int     fighters_launched;
        int     z_offs;
        int     fighter_count;
        int     torp_rx_count;
        int     torp_count;
        int     number;

        // Statistics stuff
        int     min_fighters_aboard;
        int     torps_hit;
    };
    /** Configuration for this ship, constant during fight. */
    struct Config {
        int32   StandoffDistance;
        int     BayRechargeRate;
        int     BeamRechargeRate;
        int     BeamHitOdds;
        int     TubeRechargeRate;
        int     TorpHitOdds;
        // MaxFightersLaunched is in data
        int     ShieldKillScaling;
        int     ShieldDamageScaling;
        int     CrewKillScaling;
        int     HullDamageScaling;
        int     StrikesPerFighter;
        int     FighterMovementSpeed;
        int     FighterBeamKill;
        int     FighterBeamExplosive;
        int     BayLaunchInterval;
        int32   TorpFiringRange;
        int     BeamHitFighterCharge;
        int     BeamHitShipCharge;
        int32   BeamFiringRange;
        bool    FireOnAttackFighters;
        int     FighterKillOdds;
        int     FighterFiringRange;
    };

    /** Ship data: all data as supplied by the host. */
    const TFlakShip   data;
    /** Running status information, not supplied by host. */
    Status            status;
    /** Configuration, fixed, not supplied by host. */
    Config            config;

    FlakFleet*  fleet_link;
    FlakBattle* battle_link;

    FlakShip(FlakFleet* fleet_link, FlakBattle* battle_link,
             const TFlakShip& data);
    ~FlakShip();

    void init(int number);
    void recharge();

    /** Check whether this unit is alive. */
    bool isAlive() const
        { return status.is_alive; }
    /** Check whether this unit is armed. */
    bool isArmed() const
        {
            return data.beam_count || (data.torp_lcount && data.torp_count)
                || (data.bay_count && (status.fighters_launched || data.fighter_count));
        }
    /** Check whether this unit is a planet. If not, it's a ship. */
    bool isPlanet() const
        { return data.flags & flak_IsPlanet; }
    FlakPos getPos() const;
};

//! Fleet data.
class FlakFleet {
 public:
    /** Running status, variable. */
    struct Status {
        FlakShip*  enemy_ptr;
        bool       alive;
        FlakPos    position;
    };
    /** This fleet's data. Constant through the fight, but cannot be
        \c const because we modify it when building the setup. */
    TFlakFleet data;
    /** Running status. */
    Status status;
    /** New position after movement. Only valid during
        position-recompute phase. */
    FlakPos new_position;

    FlakFleet(const TFlakFleet& data);

    /** Check whether this fleet is still alive. */
    bool isAlive() const
        { return status.alive; }
    void init();
};

/** Get ship's position. */
inline FlakPos
FlakShip::getPos() const
{
    FlakPos pos = fleet_link->status.position;
    pos.z += status.z_offs;
    return pos;
}

//! Object in space (fighter or torpedo).
class FlakObject {
 public:
    /// Kind of object.
    enum Kind {
        oFighter,               ///< Fighter.
        oTorpedo,               ///< Torpedo.
        oDeleteMe               ///< Scheduled for deletion.
    };
    Kind       kind;                    ///< Object kind.
    bool       can_change_enemy;        ///< True if this object can change its enemy (applicable to fighters).
    FlakPos    position;                ///< Current position.
    FlakShip*  enemy_ptr;               ///< Enemy ship (our target).
    FlakShip*  owner_ptr;               ///< Owner ship (who launched us).
    int        strikes;                 ///< Strikes left. For torps, nonzero if it hits, zero if it misses.
    int        kill, expl, death_flag;  ///< Weapon characteristics.
    int        speed;                   ///< Movement speed.
    uint32     vis_data;                ///< Data for use by visualizer.

    FlakObject* next;                   ///< Next in chain.

    int32 moveTowards(FlakPos pos);

    static FlakObject* cloneList(FlakObject* p);
    static void        freeList(FlakObject* p);
};

//! Information about a single player.
class FlakPlayer {
 public:
    int16        number;          ///< Player number.
    int16        num_live_ships;  ///< Number of living ships this player has.
    int32        sum_strength;    ///< Total strength, for compensation rating.
    FlakObject*  stuff;           ///< Active fighters and torps.
    bool         have_any_fighters; ///< True iff this player has (had) any fighters out.

    FlakPlayer();
    FlakPlayer(const FlakPlayer& other);
    ~FlakPlayer();
    FlakPlayer& operator=(const FlakPlayer&);

    void doPlayerGC();
    void addObject(FlakObject* obj);
    void init(int number);
};

//! Status token.
class FlakStatusToken {
    std::vector<FlakFleet::Status*>  fleets;
    std::vector<FlakShip::Status*>   ships;
    FlakPlayer                       player_status[FLAK_NUM_OWNERS];
    uint32                           seed;
    uint32                           time;
    bool                             is_terminated;

    FlakStatusToken& operator=(const FlakStatusToken&);
 public:
    FlakStatusToken(const FlakBattle& battle);
    FlakStatusToken(const FlakStatusToken& token);
    ~FlakStatusToken();
    void storeTo(FlakBattle& battle) const;
};

//! FLAK Battle Player.
class FlakBattle {
    friend class FlakPlayer;
    friend class FlakObject;
    friend class FlakStatusToken;
 public:
    /** Type for a ship number. */
    typedef unsigned int ship_t;
    /** Type for a fleet number. */
    typedef unsigned int fleet_t;

    static const ship_t no_ship = ship_t(-1);
    static const fleet_t no_fleet = fleet_t(-1);
 private:
    std::vector<int16>       attack_list;
    std::vector<FlakFleet*>  fleets;
    std::vector<FlakShip*>   ships;
    FlakPlayer               player_status[FLAK_NUM_OWNERS];
    int                      player_index[FLAK_NUM_OWNERS];
    int                      player_count;

    uint32                   seed;
    uint32                   time;
    uint32                   orig_seed;
    bool                     is_terminated;

    FlakVisualizer*          vis;

    static void    computeTorpLimit(FlakShip& attacker, const FlakShip& ship, int num_torpers);
    void           hitShipWith(FlakShip& sh, const FlakShip& firing_ship, int damage, int kill, int death_flag);

    /// \name Combat phases
    //@{
    void    chooseEnemy(FlakFleet& fleet);
    void    launchFighters(const FlakFleet& fleet);
    void    fireTorps(const FlakFleet& fleet);
    void    fireBeams(const FlakFleet& fleet);
    bool    endCheck() const;
    void    computeNewPosition(FlakFleet& fleet);
    void    doFleetGC(FlakFleet& fleet);
    void    fighterIntercept(FlakPlayer& a, FlakPlayer& b);
    bool    tryIntercept(FlakObject& pa, FlakObject& pb);
    void    fightersFire(const FlakPlayer& player);
    void    findNewBase(const FlakPlayer& player, FlakObject& fighter);
    void    moveStuff(FlakPlayer& player);
    //@}

    /// \name Initialisation
    //@{
    void    removePassiveObjects();
    void    computeInitialPositions();
    void    computeFleetSpeeds();
    void    assignInitialPositions(int player, double angle, int32 dist);
    void    adjustStrengths(int adj_to);
    //@}
 public:
    FlakBattle();
    FlakBattle(const FlakBattle& b);
    ~FlakBattle();

    /// \name Creating a fight from scratch
    //@{
    fleet_t addFleet(int player);
    ship_t  addShip(const TFlakShip& ship);
    int32   addAttackListEntry(int16 ship_id, int16 rating_bonus);
    int32   getAttackListNumber() const
        { return attack_list.size()/2; }
    void    initAfterSetup();
    void    setInitialSeed(uint32 seed);
    static void initShip(TFlakShip& ship);

    const FlakShip* findCaptor(const FlakShip& victim);
    //@}

    /// \name Loading/Saving a fight in a file
    //@{
    void    saveToStream(Buffer<char>& s, int x, int y, int32 ambient_flags);
    void    loadFromStream(string_t name, const Buffer<char>& s, TFlakBattle& header_out);
    //@}

    /* Playback */
    void    init();
    bool    playOneTick();
    void    printCheckpoint(std::ostream& os);

    /** Get ship by number.
        \param number ship number, [0, getNumShips()). */
    FlakShip& getShipByNumber(int number)
        { return *ships[number]; }
    const FlakShip& getShipByNumber(int number) const
        { return *ships[number]; }
    FlakShip* getShipById(int id, bool want_planet);
    /** Get fleet by number.
        \param number fleet number, [0, getNumFleets()). */
    FlakFleet& getFleetByNumber(int number)
        { return *fleets[number]; }
    const FlakFleet& getFleetByNumber(int number) const
        { return *fleets[number]; }
    /** Get objects owned by a player.
        \param number player number, [1, FLAK_NUM_OWNERS]. */
    FlakObject* getPlayerObjects(int number)
        { return player_status[number-1].stuff; }
    /** Get number of fleets. */
    int getNumFleets() const
        { return fleets.size(); }
    /** Get number of ships. */
    int getNumShips() const
        { return ships.size(); }
    /** Get current time. */
    uint32 getTime() const
        { return time; }
    /** Check whether we are done. */
    bool isTerminated() const
        { return is_terminated; }

    int random(int max)
        {
            seed = 0x8088405U * seed + 1;
            return ((seed >> 16) * max) >> 16;
        }

    uint32 getOrigSeed() const
        { return orig_seed; }
    void setOrigSeed(uint32 val)
        { orig_seed = val; }

    FlakVisualizer* getVisualizer() const
        { return vis; }
    void setVisualizer(FlakVisualizer* vis)
        { this->vis = vis; }
};

//! Visualisation of a fight.
class FlakVisualizer : public Object {
 protected:
    FlakBattle& the_battle;
 public:
    typedef FlakBattle::ship_t ship_t;      ///< Type containing a ship number.
    typedef FlakBattle::fleet_t fleet_t;    ///< Type containing a fleet number.

    FlakVisualizer(FlakBattle& the_battle);
    ~FlakVisualizer();

    /** Create an object (torp/fighter).
        \param obj the object, fully initialized
        \param source who is creating the object */
    virtual void createObject(FlakObject& obj, ship_t source) = 0;
    /** Destroy an object.
        - if obj is a torpedo: if violent=true, the torpedo hits.
          Its enemy_ptr points to the ship being hit; the destroyObject()
          is immediately followed by a hitShip(). enemy_ptr may also be
          null if the torp would have hit but its target is already dead.
          If violent=false, the torpedo misses.
        - if obj is a fighter: if violent=true, the fighter explodes for
          one of these reasons: an enemy beam (fighter intercept, enemy
          ship; in both cases, the destroyObject() is immediately preceded
          by fireBeam()), or because it wants to return to its base but
          none still existed (no preceding call here!). If violent is
          false, the fighter returns to a baseship.
        \param obj the object being destroyed (still intact)
        \param violent true if object is destroyed violently (torp hits, fighter killed). */
    virtual void destroyObject(FlakObject& obj, bool violent) = 0;
    /** Destroy ship. Called at the end of a tick, when death is declared. */
    virtual void destroyShip(ship_t unit) = 0;
    /** Fire a beam between two fighters. The beam always hits, obj2
        dies. The respective destroyObject(obj2, true) will be called
        immediately after. */
    virtual void fireBeam(FlakObject& obj1, FlakObject& obj2) = 0;
    /** Fire a beam from fighter at ship. The beam always hits, the
        call is immediately followed by a hitShip() call. */
    virtual void fireBeam(FlakObject& obj1, ship_t unit2) = 0;
    /** Fire beam from ship at fighter. If the beam hits, obj2 dies, and
        destroyObject(obj2, true) will be called immediately after. */
    virtual void fireBeam(ship_t unit1, FlakObject& obj2, bool hits) = 0;
    /** Fire beam from ship at ship. If the beam hits, the call is
        immediately followed by a hitShip() call. */
    virtual void fireBeam(ship_t unit1, ship_t unit2, bool hits) = 0;
    /** Hit a ship. This is called immerdiately after a weapon hit on a
        ship (fireBeam() for a beam from ship or fighter, destroyObject()
        for a torpedo)
        \param unit ship that got hit
        \param expl,kill weapon parameters
        \param death_flag zero if weapon is death ray */
    virtual void hitShip(ship_t unit, int expl, int kill, int death_flag) = 0;

    /** Get access to the battle we're playing. */
    FlakBattle& getBattle()
        { return the_battle; }
};

//! Null Visualizer.
class FlakNullVisualizer : public FlakVisualizer {
 public:
    FlakNullVisualizer(FlakBattle& the_battle)
        : FlakVisualizer(the_battle) { }

    void createObject(FlakObject& obj, ship_t source);
    void destroyObject(FlakObject& obj, bool violent);
    void destroyShip(ship_t unit);
    void fireBeam(FlakObject& obj1, FlakObject& obj2);
    void fireBeam(FlakObject& obj1, ship_t unit2);
    void fireBeam(ship_t unit1, FlakObject& obj2, bool hits);
    void fireBeam(ship_t unit1, ship_t unit2, bool hits);
    void hitShip(ship_t unit, int expl, int kill, int death_flag);
};

#endif
