/**
  *  \file game/cost.cc
  *  \brief Cost/Resource Set
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ostream>
#include "game/cost.h"
#include "cpluslib/format.h"
#include "game/cspec.h"

/** Constructor. Construct empty (zero) cost. */
GCost::GCost()
    : t(0), d(0), m(0), mc(0), sup(0)
{ }

/** Constructor. Construct from TCost structure, as stored in
    specification file. */
GCost::GCost(const TCost& cost)
    : t(cost.t), d(cost.d), m(cost.m), mc(cost.mc), sup(0)
{ }

/** Try to sell supplies to cover a money shortage. */
void
GCost::trySupplySale()
{
    if (mc < 0) {
        sup += mc;
        mc = 0;
        if (sup < 0) {
            // not enough supplies - leave it a money shortage
            mc = sup;
            sup = 0;
        }
    }
}

/** Add costs.
    \param other Value to add to this one.
    \pre other.isValid()
    \return *this */
GCost&
GCost::operator+=(const GCost& other)
{
    t += other.t;
    d += other.d;
    m += other.m;
    mc += other.mc;
    sup += other.sup;
    return *this;
}

/** Multiply in-place.
    \param n Factor to multiply this with.
    \return *this  */
GCost&
GCost::operator*=(int n)
{
    t *= n;
    d *= n;
    m *= n;
    mc *= n;
    sup *= n;
    return *this;
}

/** Multiply.
    \param n Factor to multiply with
    \return scaled cost */
GCost
GCost::operator*(int n) const
{
    GCost result(*this);
    result *= n;
    return result;
}

/** Compare for equality.
    \param other other value
    \return true iff these two costs are equal */
bool
GCost::operator==(const GCost& other) const
{
    return t == other.t
        && d == other.d
        && m == other.m
        && mc == other.mc
        && sup == other.sup;
}

/** Subtract. This automatically handles supply-sale.
    \param other Value to subtract from this one
    \pre other.isValid() */
GCost&
GCost::operator-=(const GCost& other)
{
    t -= other.t;
    d -= other.d;
    m -= other.m;
    mc -= other.mc;
    sup -= other.sup;
    trySupplySale();
    return *this;
}

/** Limit amount of items to build. This object represents the available
    resources. We attempt to build \c amount items that cost \c item each.
    \param amount Number of items to build
    \param item Cost of each item.
    \return largest n <= amount such that isEnoughFor(item*n). */
int32
GCost::limitAmount(int32 amount, const GCost& item) const
{
    if (!isValid() || amount < 0) {
        /* This cost is not valid, or amount is negative. Neither allows
           us to fulfil our postcondition, so return 0. */
        amount = 0;
    } else {
        /* Regular path */
        if (item.t != 0)
            amount = std::min(amount, t / item.t);
        if (item.d != 0)
            amount = std::min(amount, d / item.d);
        if (item.m != 0)
            amount = std::min(amount, m / item.m);
        if (item.sup != 0)
            amount = std::min(amount, sup / item.sup);
        if (item.sup + item.mc != 0)
            amount = std::min(amount, (sup+mc) / (item.sup + item.mc));
    }
    return amount;
}

/** Set component.
    \param type Component to set. Must be el_Tritanium, el_Duranium,
                el_Molybdenum, el_Money, el_Supplies.
    \param n Value to set component to */
void
GCost::set(GCargoType type, int32 n)
{
    switch (type) {
     case el_Tritanium:  t = n; break;
     case el_Duranium:   d = n; break;
     case el_Molybdenum: m = n; break;
     case el_Money:     mc = n; break;
     case el_Supplies: sup = n; break;
     default: ASSERT(0); // FIXME?
    }
}

/** Get component.
    \param type component to query
    \return amount of component (negative if missing); 0 if type is outside
    valid range */
int32
GCost::get(GCargoType type) const
{
    switch (type) {
     case el_Tritanium: return t;
     case el_Duranium: return d;
     case el_Molybdenum: return m;
     case el_Money: return mc;
     case el_Supplies: return sup;
     default: return 0;
    }
}

/** Add component.
    \param type component to add.
    \param n increment */
void
GCost::add(GCargoType type, int32 n)
{
    set(type, get(type) + n);
}

/** Remove component.
    \param type component to remove.
    \param n decrement */
void
GCost::remove(GCargoType type, int32 n)
{
    set(type, get(type) - n);
}

/** Check whether this resource amount is large enough to buy an item.
    \param other price of the item
    \return true iff this cost is equal or higher than \c other. Supply sale
    is handled. */
bool
GCost::isEnoughFor(const GCost& other) const
{
    GCost tmp(*this);
    tmp -= other;
    return tmp.isValid();
}

/** Check whether this cost is valid.
    \return true iff all components are >= 0 */
bool
GCost::isValid() const
{
    return t >= 0 && d >= 0 && m >= 0 && mc >= 0 && sup >= 0;
}

/** Check whether this cost is empty.
    \return true iff all components are zero */
bool
GCost::isEmpty() const
{
    return t == 0 && d == 0 && m == 0 && mc == 0 && sup == 0;
}

/** Convert to PHost-style string. Generates a list of words of the form "Tnnn"
    with T identifying the resource, nnn the amount. If this cost is valid,
    the resultant string will be a valid parameter for, say, PHost's "beamup"
    command.
    \return string, never blank */
string_t
GCost::toPHostString() const
{
    return GCargoSpec(*this).toPHostString();
}

/** Convert to CCScript-style string. Generates a list of words of the form
    "nnnT", with T identifying the resource (may be more than one), nnn being
    the amount. If this cost is valid, the resultant string will be a valid
    parameter for CCScript cargo operations.
    \return string, possibly blank */
string_t
GCost::toCargospec() const
{
    return GCargoSpec(*this).toCargoSpecString();
}

/** Parse a string into a GCost structure. This is intended to be used
    by the script and config interface. It accepts both PCC cargospecs
    ("nnnT") and PHost cost format ("Tnnn").

    This routine never fails; if it finds an invalid character it
    returns the cost parsed so far. */
GCost
GCost::fromString(const string_t value)
{
    return GCargoSpec(value, true).toCost();
}

/** Output GCost onto C++ stream. */
std::ostream&
operator<<(std::ostream& os, const GCost& c)
{
    return os << c.toPHostString();
}
