/**
  *  \file int/array.cc
  *  \brief Interpreter: Arrays
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/array.h"
#include "int/if/ifutil.h"
#include "int/args.h"
#include "int/interror.h"
#include "int/vmio.h"

/** Maximum total size of an array. We must define an upper limit to
    avoid getting into regions where operator new fails too easily. By
    defining the maximum size as 10001-squared, we allow an array
    indexed by two minefield Ids (the highest-possible Id sequence),
    while having a maximum memory consumption of the IntDataSegment of
    about 400 megs, 1/10 of the 32-bit address space. */
static const uint32_t MAX_ARRAY_ELEMS = 10001UL * 10001UL;


/** Move part of an array.
    This is a recursive operation.
    For a two-dimensional array, level=0 copies the whole array,
    level=1 copies a row, and level=2 copies a cell.

    \param from        [in/out] Source array. Parts of it will be destroyed.
    \param to          [in/out] Target array. Must be blank.
    \param fromDim     [in] Dimensions of source array.
    \param toDim       [in] Dimensions of target array.
    \param level       [in] Dimension counter for recursion.
    \param fromAddress [in] First address to read.
    \param toAddress   [in] First address to write. */
static void
moveArray(IntDataSegment& from,
          IntDataSegment& to,
          const std::vector<uint32_t>& fromDim,
          const std::vector<uint32_t>& toDim,
          const uint32_t level,
          uint32_t fromAddress,
          uint32_t toAddress)
{
    if (fromAddress >= from.size()) {
        // Source address is beyond from's size; nothing to do.
    } else if (level < fromDim.size()) {
        // A row
        uint32_t fromSize = 1, toSize = 1;
        for (uint32_t i = level+1; i < fromDim.size(); ++i) {
            fromSize *= fromDim[i];
            toSize *= toDim[i];
        }
        for (uint32_t i = 0; i < fromDim[level] && i < toDim[level]; ++i) {
            moveArray(from, to, fromDim, toDim, level+1, fromAddress, toAddress);
            fromAddress += fromSize;
            toAddress += toSize;
        }
    } else {
        // A cell
        from.swap(fromAddress, to, toAddress);
    }
}

/****************************** IntArrayData *****************************/

/** Constructor. Makes a blank array with no dimensions. */
IntArrayData::IntArrayData()
    : total_size(1),
      dimensions(),
      content()
{ }

/** Destructor. */
IntArrayData::~IntArrayData()
{ }

/** Add a dimension.
    \param n New dimension
    \return true if dimension accepted, false on error */
bool
IntArrayData::addDimension(int32_t n)
{
    /* We must allow dimension 0 to allow empty arrays, so we cannot
       have just one single "n < 0" test. */
    if (n < 0)
        return false;

    if (n > 0 && total_size > 0 && uint32_t(n) > MAX_ARRAY_ELEMS / total_size)
        return false;

    dimensions.push_back(n);
    total_size *= n;
    return true;
}

/** Compute index for this array.
    \param args   [in/out] Argument block containing user-supplied indices
    \param result [out] Linear index
    \retval true Result has been produced
    \retval false One of the indices was null
    \throw IntError One of the indices was out-of-range */
bool
IntArrayData::computeIndex(IntArgBlock& args, uint32_t& result) const
{
    args.checkArgumentCount(dimensions.size());
    result = 0;
    for (uint32_t i = 0; i < dimensions.size(); ++i) {
        // Read arg. Check for null.
        int32_t me;
        if (!checkIntArg(me, args.getNext()))
            return false;

        // Range check.
        if (me < 0 || uint32_t(me) >= dimensions[i])
            throw IntError::rangeError();

        // Compute index. Horner schema.
        result = result * dimensions[i] + me;
    }
    return true;
}

/** Resize array. Updates the array in-place, keeping the values if possible.
    \param tpl Template for new array. The new array will have the same dimensions as this one. */
void
IntArrayData::resize(const IntArrayData& tpl)
{
    // Verify number of dimensions
    checkArgumentCount(tpl.getNumDimensions(), dimensions.size(), dimensions.size());

    // Can we resize in-place?
    bool allowInPlace = true;
    for (uint32_t i = 1; i < dimensions.size(); ++i) {
        if (tpl.dimensions[i] != dimensions[i]) {
            allowInPlace = false;
            break;
        }
    }

    if (allowInPlace) {
        // In-place resize. This means just discard excess elements, if any.
        if (content.size() > tpl.total_size) {
            content.pop(content.size() - tpl.total_size);
        }
    } else {
        // Out-of-place resize. Slow.
        IntDataSegment newData;
        moveArray(content, newData, dimensions, tpl.dimensions, 0, 0, 0);
        content.swap(newData);
    }

    // Copy new metadata
    total_size = tpl.total_size;
    dimensions = tpl.dimensions;
}

/******************************** IntArray *******************************/

/** Constructor.
    \param data Our array */
IntArray::IntArray(Ptr<IntArrayData> data)
    : data(data)
{ }

// IntIndexableValue:
IntValue*
IntArray::get(IntArgBlock& args)
{
    uint32_t index;
    if (data->computeIndex(args, index)) {
        return IntValue::cloneOf(data->content[index]);
    } else {
        return 0;
    }
}

void
IntArray::set(IntArgBlock& args, IntValue* value)
{
    uint32_t index;
    if (data->computeIndex(args, index)) {
        data->content.set(index, value);
    } else {
        throw IntError::typeError(IntError::tyInteger);
    }
}

int32_t
IntArray::getDimension(int32_t dim)
{
    if (dim == 0) {
        return data->getNumDimensions();
    } else {
        return data->getDimension(dim-1);
    }
}

IntContext*
IntArray::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntArray::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntArray::store(IntTagNode& tag, Stream& /*aux*/)
{
    IntVMSaveContext* vsc = IntVMSaveContext::getCurrentInstance();
    if (vsc != 0) {
        tag.tag   = IntTagNode::Tag_Array;
        tag.value = vsc->addArray(*data);
    } else {
        throw IntError::notSerializable();
    }
}

IntValue*
IntArray::clone()
{
    return new IntArray(data);
}
