/**
  *  \file int/keymap.h
  *  \brief Interpreter: Keymaps
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_KEYMAP_H
#define PCC_V2_INT_KEYMAP_H

#include <vector>
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "int/value.h"

class IntKeymap;

/** Keymap reference. User code dealing with keymaps uses objects of type IntKeymapRef.
    This might one day be a smart pointer. Currently, all live IntKeymap's are managed
    by the keymap.cc module in a ptr_vector, so there are no lifetime issues. */
typedef IntKeymap* IntKeymapRef;

/** Keymap. A keymap maps keys to command/condition pairs. Keys are specified as uint32_t's
    (see UIKey). Commands and conditions are also specified as uint32_t's, which must be
    atoms that map to a statement or expression, respectively.

    Conditions will be used to detect whether buttons should be enabled or disabled. Space
    for them has therefore been reserved in the data structure.

    A keymap can have one or many parents, thus building an acyclic directed graph.
    Keys not found in a keymap are looked up in all its parents, from first to last. */
class IntKeymap {
    struct Entry {
        uint32_t key;
        uint32_t command;
        uint32_t condition;
        Entry(uint32_t key, uint32_t command, uint32_t condition)
            : key(key), command(command), condition(condition)
            { }
    };
    std::vector<IntKeymap*> parents;             ///< All this keymap's parents.
    std::vector<Entry> keys;                     ///< All this keymap's keys.
    string_t name;                               ///< Name of this keymap.

    IntKeymap(const string_t& name);
    const Entry* lookup(uint32_t key, IntKeymapRef* where) const;
 public:
    ~IntKeymap();

    void     addKey(uint32_t key, uint32_t command, uint32_t condition);
    void     addParent(IntKeymapRef km);
    uint32_t lookupCommand(uint32_t key) const;
    uint32_t lookupCommand(uint32_t key, IntKeymapRef& where) const;
    uint32_t lookupCondition(uint32_t key) const;
    bool     hasParent(IntKeymap* km) const;
    const string_t& getName() const;

    uint32_t     getNumDirectParents() const;
    IntKeymapRef getDirectParent(uint32_t index) const;

    static IntKeymapRef getKeymapByName(string_t name);
    static IntKeymapRef createKeymap(string_t name);
    static size_t getNumKeymaps();
    static IntKeymapRef getKeymapByIndex(size_t index);
};

/** Keymap value. Script code uses these (temporarily) to refer to keymaps. These values
    are not exposed to the user. 'Bind' and 'CreateKeymap' commands are compiled into
    fixed command sequences that convert a string into a keymap using ukeylookup or
    ukeycreate; after operating on that value using bkeyaddparent, tkeyadd, the
    IntKeymapValue will be dropped and the user will not be able to access it. */
class IntKeymapValue : public IntValue {
    IntKeymapRef keymap;
 public:
    IntKeymapValue(IntKeymapRef keymap);
    ~IntKeymapValue();
    IntKeymapRef getKeymap() const;

    string_t toString(bool readable);
    void store(IntTagNode& tag, Stream& aux);
    IntKeymapValue* clone();
};

void initKeymaps();
void doneKeymaps();
uint32_t parseKey(string_t str);
string_t formatKey(uint32_t key);
IntKeymapValue* makeKeymapValue(IntKeymapRef km);

/**************************** Inline functions ***************************/

/** Get name of a keymap. */
inline const string_t&
IntKeymap::getName() const
{
    return name;
}

/** Get number of direct parents. */
inline uint32_t
IntKeymap::getNumDirectParents() const
{
    return parents.size();
}

/** Get reference to a direct parent.
    \param index [0,getNumDirectParents()) Index */
inline IntKeymapRef
IntKeymap::getDirectParent(uint32_t index) const
{
    return parents[index];
}

/** Get keymap contained in this IntKeymapValue. */
inline IntKeymapRef
IntKeymapValue::getKeymap() const
{
    return keymap;
}

#endif
