/**
  *  \file u/t_util_helpindex.cc
  *  \brief Tests for HelpIndex
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/helpindex.h"

#include <cstring>
#include "u/t_util.h"
#include "io/streams.h"
#include "io/dirs.h"

/** Test scenarios. */
void
UtilHelpIndexTestSuite::testMulti()
{
    // Create files
    const char FILE1[] = "<help><page id=\"toc\"></page><page id=\"f1\"></page></help>";
    const char FILE2[] = "<help priority=\"99\"><page id=\"toc\"></page><page id=\"f2\"></page></help>";
    const char FILE3[] = "<help priority=\"99\"><page id=\"toc\"></page><page id=\"f2\"></page></help>";
    Ptr<Directory> dir = Directory::openDir(Directory::getWorkingDirectory());
    dir->openFile("__tmp1", Stream::C_CREATE)->writeT(FILE1, std::strlen(FILE1));
    dir->openFile("__tmp2", Stream::C_CREATE)->writeT(FILE2, std::strlen(FILE2));
    dir->openFile("__tmp3", Stream::C_CREATE)->writeT(FILE3, std::strlen(FILE3));

    // Create help index, starting with two files
    HelpIndex t;
    t.addFile("__tmp1", "o1");
    t.addFile("__tmp2", "o2");

    // Not found
    {
        std::vector<HelpIndex::Node*> out;
        t.find("qq", out);
        TS_ASSERT_EQUALS(out.size(), 0U);
    }

    // Verify toc
    {
        std::vector<HelpIndex::Node*> out;
        t.find("toc", out);
        TS_ASSERT_EQUALS(out.size(), 2U);
        TS_ASSERT(out[0] != 0);
        TS_ASSERT_EQUALS(out[0]->file.name, "__tmp2");
        TS_ASSERT(out[1] != 0);
        TS_ASSERT_EQUALS(out[1]->file.name, "__tmp1");
    }

    // Verify f2
    {
        std::vector<HelpIndex::Node*> out;
        t.find("f2", out);
        TS_ASSERT_EQUALS(out.size(), 1U);
        TS_ASSERT(out[0] != 0);
        TS_ASSERT_EQUALS(out[0]->file.name, "__tmp2");
    }

    // Add another file and verify again
    t.addFile("__tmp3", "o3");

    // Verify toc
    {
        std::vector<HelpIndex::Node*> out;
        t.find("toc", out);
        TS_ASSERT_EQUALS(out.size(), 3U);
        TS_ASSERT(out[0] != 0);
        TS_ASSERT_EQUALS(out[0]->file.name, "__tmp3");
        TS_ASSERT(out[1] != 0);
        TS_ASSERT_EQUALS(out[1]->file.name, "__tmp2");
        TS_ASSERT(out[2] != 0);
        TS_ASSERT_EQUALS(out[2]->file.name, "__tmp1");
    }

    // Verify f2
    {
        std::vector<HelpIndex::Node*> out;
        t.find("f2", out);
        TS_ASSERT_EQUALS(out.size(), 1U);
        TS_ASSERT(out[0] != 0);
        TS_ASSERT_EQUALS(out[0]->file.name, "__tmp3");
    }

    // Remove o2
    t.removeFilesByOrigin("o2");

    // Verify toc
    {
        std::vector<HelpIndex::Node*> out;
        t.find("toc", out);
        TS_ASSERT_EQUALS(out.size(), 2U);
        TS_ASSERT(out[0] != 0);
        TS_ASSERT_EQUALS(out[0]->file.name, "__tmp3");
        TS_ASSERT(out[1] != 0);
        TS_ASSERT_EQUALS(out[1]->file.name, "__tmp1");
    }

    // Remove files
    dir->erase("__tmp1");
    dir->erase("__tmp2");
    dir->erase("__tmp3");
}

/** Test adding a missing file. Must not throw. */
void
UtilHelpIndexTestSuite::testMissing()
{
    HelpIndex t;
    TS_ASSERT_THROWS_NOTHING(t.addFile("__q2poiwknskdflahuw0e298x", "o1"));

    std::vector<HelpIndex::Node*> out;
    TS_ASSERT_THROWS_NOTHING(t.find("p", out));
    TS_ASSERT(out.empty());
}
