/**
  *  \file ui/table.h
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_TABLE_H
#define PCC_V2_UI_TABLE_H

#include "cpluslib/string.h"
#include "cpluslib/object.h"
#include "ui/skin.h"

class GfxFont;
class GfxCanvas;
class UISkin;

class UIAbstractTableWriter : public Object {
    int  write_flags;           ///< Flags to pass to writeAnyCell
    int  col_width;             ///< Left-just: minimum size so far; Right-just: column width
    int  cur_line;              ///< Current line, 0-based
    int  cur_x;                 ///< Current X offset
 protected:
    enum {
        fl_Solid     = 1,
        fl_LJust     = 2,
        fl_Fixed     = 4,
        fl_Variable  = 8,
        fl_AVAILABLE = 16
    };

    int getCurrentLine() const { return cur_line; }
    int getCurrentX() const { return cur_x; }
    int getCurrentWidth() const { return col_width; }

    virtual int writeAnyCell(UISkin::Color c, const string_t& s, int flags, int col_size) = 0;
    virtual int sizeAnyCell(const string_t& s) = 0;
    void writeCell(UISkin::Color c, const string_t& s, int flags);
 public:
    UIAbstractTableWriter();

    virtual void finish();

    void sizeCell(string_t s);
    
    void writeFixedCell(UISkin::Color c, string_t s) { writeCell(c, s, fl_Fixed); }
    void writeFixedCell(UISkin::Color c, const char* s);
    void writeVariableCell(UISkin::Color c, string_t s) { writeCell(c, s, fl_Variable | fl_Solid); }
    void writeVariableCell(UISkin::Color c, const char* s);

    void newLColumn(int min_size = 0);
    void newLColumnUpTo(int end_x);
    void newRColumn(int width);

    virtual GfxPoint getCurrentPosition() const = 0;

    static const int COLUMN_SEPARATOR = 6;
};

class UITableSizer : public UIAbstractTableWriter {
    GfxFont& fnt;
    int max_lines;
 protected:
    int writeAnyCell(UISkin::Color c, const string_t& s, int flags, int col_size);
    int sizeAnyCell(const string_t& s);
 public:
    UITableSizer(int font_nr);
    UITableSizer(GfxFont& fnt);
    ~UITableSizer();

    int getWidth() const { return getCurrentX() + getCurrentWidth(); }
    int getLines() const { return max_lines; }
    int getHeight() const;

    GfxPoint getCurrentPosition() const;
};

class UITableWriter : public UIAbstractTableWriter {
    GfxContext& ctx;
    int font_nr;

    GfxRect limit;
 protected:
    int writeAnyCell(UISkin::Color c, const string_t& s, int flags, int col_size);
    int sizeAnyCell(const string_t& s);
 public:
    UITableWriter(GfxContext& ctx, GfxRect limit, int font_nr);
    ~UITableWriter();
    void finish();

    int getCurrentLine() const { return UIAbstractTableWriter::getCurrentLine(); }

    GfxPoint getCurrentPosition() const;
};

#endif
