/**
  *  \file ui/widgets/abstractbutton.h
  *  \brief UIAbstractButton class
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_WIDGETS_ABSTRACTBUTTON_H
#define PCC_V2_UI_WIDGETS_ABSTRACTBUTTON_H

#include "ui/ui.h"

/** \class UIAbstractButton
    \brief Standard Button

    This class provides the event handling mechanisms of a standard
    clickable button (checkbox, radiobutton, ...). The button can be
    have an associated hot-key if so desired.

    Buttons are triggered by a keystroke or mouse click.

    Derived classes must override draw().

    Buttons fire a command, a stop, or a keystroke. You can attach
    own handlers to sig_fired. */
class UIAbstractButton : public UIWidget {
    void callStop(int);
    void postCommand(int);

    UIKey key;

 public:
    enum {
        bf_Pressed       = st_Last*2,     //!< Button is currently pressed.
        bf_Highlight     = st_Last*4,     //!< Button is highlighted ("on").
        bf_Key           = st_Last*8,     //!< This is a "key" button. That is, it generates a keystroke and does not react upon keystrokes.
        bf_SmallSize     = st_Last*16,    //!< Use "small" font.
        bf_PressOverride = st_Last*32,    //!< Button should appear depressed.
        bf_LeftJust      = st_Last*64     //!< Left-justify text.
    };

    /* Constructors */
    UIAbstractButton();
    UIAbstractButton(UIKey key);
    UIAbstractButton(int id, UIKey key);
    ~UIAbstractButton();

    /* Button handlers */
    UIAbstractButton& addCommand();
    UIAbstractButton& addStop();
    UIAbstractButton& addKey();

    template<typename T> UIAbstractButton& addCall(T* object, void (T::*function)(int))
        { sig_fired.add(object, function); return *this; }
    template<typename T> UIAbstractButton& addCall(T* object, void (T::*function)())
        { sig_fired.add(object, function); return *this; }

    /* Event handlers */
    bool handleEvent(const UIEvent& e, bool second_pass);
    bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
    void onStateChange(int astate, bool enable);
    void fire(int arg);

    UIKey getKey() const;
    void setKey(UIKey);

    Signal1<int> sig_fired;
};

inline UIKey
UIAbstractButton::getKey() const
{
    return key;
}

inline void
UIAbstractButton::setKey(UIKey key)
{
    this->key = key;
}

#endif
